/* umtsmon - a program to control/monitor your UMTS card
 * This file copyright (C) 2006,2007,2008,2009  Klaas van Gend
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation
 * applicable version is GPL version 2 only.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "umtsmon_global.h"

#include "PPPConnection.h"
#include "Query.h"
#include "Runner.h"
#include "Profile.h"
#include "TheDeviceManagerSingleton.h"
#include "TheSettingsSingleton.h"
#include "Popup.h"
#include "SerialPort.h"
#include "FileStuff.h"

#include <qregexp.h>
#include <qstring.h>
#include <sys/types.h>
#include <signal.h>
#include <cstdlib>

// the actual definition of thePPPDState
static PPPObserverInterface::PPPDState  thePPPDState = PPPObserverInterface::STOPPED;
static pid_t theRunningPPPDPID = -1;


// the function to be called by atexit
// to make sure the PPPD daemon is shut down
static void run_at_exit(void)
{
	DEBUG5("PPPConnection run_at_exit() called\n");
	Runner myRunner;
	int myPID = myRunner.doesPPPDexist();
	if (myPID!=-1)
	{
		if (thePPPDState != PPPObserverInterface::NOT_OURS_RUNNING)
		{
			// thePPPDState is either RUNNING or STARTING
			DEBUG3("Killing PPPD - process %d\n", myPID);
			kill(myPID, SIGTERM);
			theRunningPPPDPID = -1;
		}
	}
}

static void run_at_signal(__attribute__ ((unused)) int aSignalNumber)
{
	run_at_exit();
	exit(1);
}


PPPConnection::PPPConnection(void) :
			theCIInhibitorPtr(NULL),
			hasOnlyOnePort(false),
			thePPPThreadPtr(NULL)
{
	atexit(run_at_exit);
	signal(SIGINT, run_at_signal);
	signal(SIGHUP, run_at_signal);

	// let's assume the worst
	setNewPPPState(PPPObserverInterface::NOT_OURS_RUNNING);
	isPPPConnected();

	// for some compatibility stuff, we need the exact version number of pppd
	Runner myRunner;
	myRunner.runCommand(Runner::PPPD, "--version");
	QString myTotalVersionNr = myRunner.getStdErrStringList()[0];
	int myFirst = myTotalVersionNr.find(QRegExp("\\d"), 0);
	int myLast  = myTotalVersionNr.find(QRegExp("\\d"), -1);
	thePPPVersionNumber = myTotalVersionNr.mid(myFirst,myLast-myFirst+1);
	DEBUG4("pppd version string '%s' results in '%s'.\n",
			myTotalVersionNr.ascii(),
			thePPPVersionNumber.ascii());
}


PPPConnection::~PPPConnection(void)
{
	DEBUG4("PPPConnection::~PPPConnection(void)\n");
	stopPPP();
}


void
PPPConnection::stopPPP(void)
{
	run_at_exit();
}

bool PPPConnection::isPPPStarting(void)
{
	// if PPP is running, it is not starting anymore :-)
	if (isPPPConnected())
		return false;
	return thePPPDState == PPPObserverInterface::STARTING;
}


bool
PPPConnection::isPPPConnected(void)
{
	// let's try to get a PID for a running pppd
	// without reading through all of /proc every time
	pid_t myPID = -1;
	// if the running pppd process ID still exists,
	// we do not need to check for a *new* pppd pid
	if (theRunningPPPDPID>0
		&& FileStuff::doesFileExist("/proc/"+QString::number(theRunningPPPDPID)+"/stat"))
	{
		myPID = theRunningPPPDPID;
	}
	else
	{
        // if there is a routing table set up for pppX, a pppd is running
        if (FileStuff::doesFileContainString("/proc/net/route", QString("ppp"))
        		&& thePPPDState!=PPPObserverInterface::STOPPED)
        {
			myPID = Runner::doesPPPDexist();
        }
	}

	switch (thePPPDState)
	{
	case PPPObserverInterface::ERROR:
		setNewPPPState(PPPObserverInterface::STOPPED);
		break;
	case PPPObserverInterface::STOPPED:
		if (myPID!=-1)
		{
			DEBUG2("PPPConnection::isPPPConnected someone else started a pppd\n");
			setNewPPPState(PPPObserverInterface::NOT_OURS_RUNNING);
		}
		if (theRunningPPPDPID!=-1)
		{
			// shouldn't happen
			DEBUG1("PPPConnection::isPPPConnected state changed to STOPPED "
			       "without updating theRunningPPPDPID\n");
			assert(false);
			theRunningPPPDPID=-1;
		}
		break;
	case PPPObserverInterface::STARTING:
		if (thePPPThreadPtr == NULL)
		{
			// something went wrong ?!?
			setNewPPPState(PPPObserverInterface::STOPPED);
			thePPPDStdErr.push_back(APPNAME " internal Error: not thePPPThreadPtr");
			break;
		}
		if (thePPPThreadPtr->isDone())
		{
			DEBUG2("isDone is true\n");
			// so, was it successful or not?
			if (myPID == theRunningPPPDPID)
			{
				// euh? finished *and* still running?
				// that means we have a zombie
				//    - process is dead, we have to reap it...
				myPID = -1;
			}
			if (myPID == -1)
			{
				theRunningPPPDPID=-1;
				thePPPDStdOut = thePPPThreadPtr->getStdOut();
				thePPPDStdErr = thePPPThreadPtr->getStdErr();
				if(thePPPVerbosity > 1)
				{
					QStringList::Iterator it;
					DEBUG4("*****output of PPP to stdout\n");
					for (it = thePPPDStdOut.begin(); it != thePPPDStdOut.end(); ++it )
					{
				        DEBUG4("* %s\n", (*it).ascii());
					}
					DEBUG4("*****end-of-output\n");
					DEBUG4("*****output of PPP to stderr\n");
					for (it = thePPPDStdErr.begin(); it != thePPPDStdErr.end(); ++it )
					{
				        DEBUG4("* %s\n", (*it).ascii());
					}
					DEBUG4("*****end-of-output\n");
				}
				setNewPPPState(PPPObserverInterface::ERROR);
			}
			else
			{
				theRunningPPPDPID = myPID;
				setNewPPPState(PPPObserverInterface::RUNNING);
			}
			DEBUG3("thread end - wait and cleanup\n");
			delete thePPPThreadPtr;
			thePPPThreadPtr = NULL;
		}
		else
		{
			DEBUG3("PPPDThread still running\n");
		}
		break;
	case PPPObserverInterface::RUNNING:
		if (myPID==-1 || myPID!=theRunningPPPDPID)
		{
			DEBUG2("Our PPPD was killed by someone\n");
			theRunningPPPDPID=-1;
			setNewPPPState(PPPObserverInterface::STOPPED);
		}
		break;
	case PPPObserverInterface::NOT_OURS_RUNNING:
		if (myPID==-1)
		{
			DEBUG2("Someone else's PPPD was killed\n");
			setNewPPPState(PPPObserverInterface::STOPPED);
		}
		break;
	}

	DEBUG5("The PPPDState is %d\n", thePPPDState);

	// we need the ConnectionInfoInhibitor here if we have only one serial port
	// and we're not in 'stopped'.
	// note that PPPThread and startPPP also have CIIs to make sure that we
	// don't run into simulatenous accessess.
	if (hasOnlyOnePort)
	{
		if (thePPPDState==PPPObserverInterface::STOPPED)
		{
			if (theCIInhibitorPtr!=NULL)
			{
				TheDeviceManagerSingleton::resetDevice();
				delete theCIInhibitorPtr;
				theCIInhibitorPtr=NULL;
			}
		}
		else
		{
			if (theCIInhibitorPtr==NULL)
			{
				theCIInhibitorPtr = new ConnectionInfoInhibitor;
			}
		}
	}

	if (thePPPDState==PPPObserverInterface::STOPPED
			|| thePPPDState==PPPObserverInterface::STARTING)
		return false;
	else
		return true;
}


bool
PPPConnection::startPPP()
{
	// just for this startup, create a Profile - which should be the currently selected profile
	Profile myProfile;
 
	theSerialPortName = TheDeviceManagerSingleton::getDeviceCapsRef().getPPPSerialPortName();
	assert(!theSerialPortName.isEmpty());

	// check if AT and PPP are the same device
	if (TheDeviceManagerSingleton::getDeviceCapsRef().getATSerialPortName() == theSerialPortName)
		hasOnlyOnePort = true;
	else
		hasOnlyOnePort = false;

	// just as a precaution, some sanity checks.
	// we will be checking if a PPP connection is already running.
	// This shouldn't be the case - that's why there is an assert(false).
	// however, if this happens in a release build, just return false.
	DEBUG3("PPPConnection::startPPP entry\n");
	if (isPPPConnected())
	{
		// if happens in debug build: make a lot of noise
		// if happens in release build: just error out of here
		DEBUG2("PPPConnection::startPPP, PPP is already connected ?!?\n");
		stopPPP();
		sleep(1);
		// this shouldn't happen - crash big time on debug builds
		assert(isPPPConnected()==false);
		// and a fall-through prevention for release builds
		if (isPPPConnected())
			return false;
	}

	// remove everything that might be left from previous connections
	assert(thePPPThreadPtr==NULL);
	thePPPDStdOut.clear();
	thePPPDStdErr.clear();


	// from now on until the exit of this member, prevent
	// ConnectionInfo from interfering with the AT serial port
	setNewPPPState(PPPObserverInterface::STARTING);
	assert(theCIInhibitorPtr == NULL);
	if (hasOnlyOnePort)
		theCIInhibitorPtr = new ConnectionInfoInhibitor();

	// we cannot connect if this is not set
	if ( theSerialPortName .isEmpty())
	{
		DEBUG2("PPPConnection::startPPP - no PPP device ?!?\n");
		return false;
	}

	// let's do the CHATting :-)
	if (!chat(myProfile))
	{
		DEBUG2("PPPConnection::startPPP - chatting failed ?!?\n");
		return false;
	}

	// let's assemble the command line for pppd and send it to the logs
	QStringList myPPPArguments = assembleArgumentList(myProfile);
	QStringList::const_iterator i = myPPPArguments.begin();
	DEBUG4("Assembled PPP string is: '");
	for (; i != myPPPArguments.end(); i++)
		DEBUG4NT("%s ", (*i).ascii());
	DEBUG4NT("'\n");

	setNewPPPState(PPPObserverInterface::STARTING);
	thePPPThreadPtr = new PPPThread(myPPPArguments);
	thePPPThreadPtr->start();
	// if there is a PPP running now, it's ours.
	// but it might take a while before all scheduling is done,
	// so let's wait for it...
	theRunningPPPDPID = -1;
	for (int i=0; i<5 && theRunningPPPDPID==-1; i++)
	{
		DEBUG5("Waiting for ppp PID\n");
		sleep(1);
		theRunningPPPDPID = Runner::doesPPPDexist();
		DEBUG5("Found PPP PID %d\n", theRunningPPPDPID);
	};
	return true;
}

bool PPPConnection::chat(const Profile& aProfile) const
{
	// have to declare all variables here because I want to use goto...
	bool myResult = false;
	unsigned int myCID;
	QString myConnectStr;
	QString myAPN;
	
	DEBUG3("PPPConnection::chat() start\n");
	// most devices want to have CHAT happen on the PPP port, not the AT port
	// 
	// let's create a new serial port, that will temporarily
	// (as long as this temp class exists) replace the original serial
	// port for queries. As this temp object will be deleted upon exit of this
	// member, the original state will be returned automatically
	TempSerialPort* myTempPortPtr = NULL;
	if (TheDeviceManagerSingleton::getDeviceCapsRef().getChatSerialPortName() == theSerialPortName)
	{
		DEBUG3("CHAT is supposed to go over the PPP port.\n");
		myTempPortPtr = new TempSerialPort();
		if (myTempPortPtr->openDev(theSerialPortName)==false)
		{
			DEBUG2("PPPConnection::startPPP - cannot open port '%s' ?!?\n", theSerialPortName.ascii());
			goto bad_end;
		}
	}
	else
	{
		DEBUG3("CHAT is supposed to go over the AT port - no modifications to Query required.\n");
	}

	// AT&F: reset modem to default profile
	// FIXME: if this fails, we should do pccardctl eject and insert and retry
	if (sendChatString("AT&F")==false)
		goto bad_end;

	// ATE1: enable echo
	// FIXME: if this fails, we should do pccardctl eject and insert and retry
	if (sendChatString("ATE1")==false)
		goto bad_end;

	// build up the connectionstring
	// like "AT+CGDCONT=1,\"IP\",\"office.vodafone.nl\""
	//
	// security check: anAPN should not contain any quotes or spaces
	myAPN = aProfile.getAPN();
	if ( myAPN.contains("\"") || myAPN.contains("\"") ||
	     myAPN.contains("'") || myAPN.contains(" ") )
	{
		Popup::Warning(QObject::tr("Your APN contains illegal characters.\nPlease go back to your Profile page and change the APN."));
		DEBUG1("ERROR: APN contains illegal character\n");
		goto bad_end;
	}
	if (myAPN.isEmpty())
	{
		Popup::WarningWithMemory("NoAPNMsg", QObject::tr("Warning: You did not specify an 'APN' in your profile.\nThis may work ok, it may not - depending on your network operator."));
	}

	// myCID needs to be adjustable: different number for cell phones on USB or BT
	myCID = aProfile.getCID();
	myConnectStr = "AT+CGDCONT=";
		myConnectStr += QString::number(myCID);
		myConnectStr += ",\"IP\",\"" + myAPN + "\"";

	if (sendChatString(myConnectStr)==false)
		goto bad_end;

 	{
 	QString myDialStr = "";
 	if ( aProfile.getCIDPrefix().stripWhiteSpace().isEmpty() )
 		myDialStr = "ATD*99***" + myConnectStr.number(myCID) + "#";
	else
 		myDialStr = "ATD" + aProfile.getCIDPrefix() + QString::number(myCID) + "#";

 	DEBUG3("Dial String is: %s \n", myDialStr.ascii() );
 	Query myQuery(myDialStr);
	// don't expect an OK following this one...
 	// however, we much catch an "ERROR"
	if (myQuery.run()== Query::ERROR)
	{
		DEBUG1("ATD returned ERROR - connection failed\n");
		goto bad_end;
	}
 	}
 	myResult = true;

 bad_end:
	DEBUG3("PPPConnection::chat() end\n");
	if (myTempPortPtr)
	{
		delete myTempPortPtr;
		myTempPortPtr = NULL;
	}
	return myResult;
}


bool 
PPPConnection::sendChatString(const QString& aChatString) const
{
	Query myQuery(aChatString);
	if (myQuery.runUntilDone() != Query::OK)
		return false;
	return true;
}


QStringList PPPConnection::assembleArgumentList(const Profile& aProfile) const
{
  	QStringList myPPPArguments;
  	myPPPArguments.push_back("idle");   myPPPArguments.push_back("7200");
	if (aProfile.getNoAuth())
		myPPPArguments.push_back("noauth");
  	myPPPArguments.push_back("asyncmap"); myPPPArguments.push_back("0");

	// PPP related stuff
	// we only ask PPP to detach if connection is successfully up
	myPPPArguments.push_back("updetach");
	if (thePPPVerbosity > 0)
	{
		myPPPArguments.push_back("dump");
		myPPPArguments.push_back("debug");
		myPPPArguments.push_back("debug");
		myPPPArguments.push_back("debug");
	}

	// modem-related stuff
	myPPPArguments.push_back("460800");
	myPPPArguments.push_back("lock");
	myPPPArguments.push_back("crtscts");
	myPPPArguments.push_back("modem");
	myPPPArguments.push_back(theSerialPortName);

	// header compressions and such
	myPPPArguments.push_back("noipx");
//	if (aProfile.getNoCompression())
	if (true)
	{
		myPPPArguments.push_back("novj");
	}
	if (aProfile.getNoCCP())
		myPPPArguments.push_back("noccp");

	// routing related stuff
	if (aProfile.getIsDefaultRoute())
		myPPPArguments.push_back("defaultroute");
	if (aProfile.getReplaceDefaultRoute())
		myPPPArguments.push_back("replacedefaultroute");
	if (aProfile.getUsePeerDns())
		myPPPArguments.push_back("usepeerdns");
	if (aProfile.getNoIpDefault())
		myPPPArguments.push_back("noipdefault");

	// login info
	if (!aProfile.getUsername().isEmpty())
	{
		myPPPArguments.push_back("user");
		myPPPArguments.push_back(aProfile.getUsername());
	}
	if (!aProfile.getPassword().isEmpty())
	{
		myPPPArguments.push_back("password");
		myPPPArguments.push_back(aProfile.getPassword());
	}

	return myPPPArguments;
}




void PPPConnection::setNewPPPState(PPPObserverInterface::PPPDState aNewState)
{
	// only go through the burden of notify if state really changed
	if (aNewState != thePPPDState)
	{
		DEBUG3("PPPConnection::setNewPPPState state %d to %d\n", thePPPDState, aNewState);
		thePPPDState = aNewState;
		PPPObserverInterface* myObserverPtr = firstObserver();
		while (myObserverPtr != NULL)
		{
			notifyObserver(myObserverPtr);
			myObserverPtr = nextObserver();
		}
	}
}

void PPPConnection::notifyObserver(PPPObserverInterface* myObserverPtr) const
{
	myObserverPtr->newPPPState(thePPPDState);
}

PPPConnection::PPPThread::PPPThread(const QStringList& anArgumentList)
	: QThread(), theArgumentList(anArgumentList), theReturnValue(-1), isRunning(false)
{
	DEBUG2("PPPConnection::PPPThread::PPPThread()\n");
}


void PPPConnection::PPPThread::run()
{
	DEBUG2("PPPThread::run, before runCommand\n");
	theReturnValue = theRunner.runCommand(Runner::PPPD, theArgumentList);
	DEBUG2("PPPThread::run, after runCommand\n");
	// end of the thread - we're done
	isRunning = false;
}
