/* umtsmon - a program to control/monitor your UMTS card
 * Copyright (C) 2006,2007,2008  Klaas van Gend
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation
 * applicable version is GPL version 2 only.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef PPPCONNECTION_H_
#define PPPCONNECTION_H_

#include "Runner.h"
#include "ObserverManager.h"
#include "ConnectionInfo.h"
#include <qstring.h>
#include <qstringlist.h>
#include <qthread.h>

// forward declarations
class Profile;
class DeviceCapabilities;

/** implement this interface and attach yourself to PPPConnection if you want
 *  to hear when the PPPState changes
 */
class PPPObserverInterface
{
public:
	enum PPPDState
	{
		STOPPED,
		STARTING,
		RUNNING,
		NOT_OURS_RUNNING,
		ERROR
	};

	/** implement this and attach to PPPConnection to get info
	 *  about state changes of PPPConnection
	 */
	virtual void newPPPState(PPPDState aNewState) = 0;
	
	/// required but not used: virtual destructor
	virtual ~PPPObserverInterface() {};
};



/** The PPPConnection class initiates, monitors and manages the PPP daemon and
 *  the connection it makes. 
 */
class PPPConnection :  
		public ObserverManager<PPPObserverInterface>
{
public:

	/// constructor
	PPPConnection(void);

	/** This member retrieves some info from the Device
	 *  @param: aDeviceCapabilitiesPtr pointer to retrieve PPP portname from
	 */
	void setDeviceInfo(const DeviceCapabilities* aDeviceCapabilitiesPtr);
	 
	/// destructor
	virtual ~PPPConnection(void);
	
	/** starts the PPP Connection
	 *  @returns true if succesful
	 */
	bool startPPP();
	
	/// stops the PPP Connection, does nothing if there is none
	void stopPPP(void);

	/// @returns a string containing the pppd version, e.g. "2.4.4"
	const QString& getPPPVersion(void)
		{ return thePPPVersionNumber; };

	bool isPPPStarting();
	bool hasErrors() {return thePPPDStdErr.count()!=0;};
	
	ConnectionInfoInhibitor* theCIInhibitorPtr;

	/// stringlist with all the stdout of the most recent PPP run
	QStringList thePPPDStdOut;
	/// stringlist with all the stderr of the most recent PPP run
	QStringList thePPPDStdErr;

	/** returns true if a PPP daemon is running
	 *  note that it is not very sharp - it will return true if ANY
	 *  pppd is running, not just if *our* pppd is running
	 */
	bool isPPPConnected(void);

private:
	/** implementation of the pure virtual member in ObserverManager;
	 *  the implementation should perform the notification of the observer pointed to.
	 */
	virtual void notifyObserver(PPPObserverInterface* anObserver) const;

	/** cache the serial port to do the PPP connection on
	 *  Note that we will still use the AT port for the dialling AT commands
	 */	
	QString theSerialPortName;
	/// true if AT and PPP serial ports are the same device
	bool hasOnlyOnePort;

	/// a string containing the PPP version
	QString thePPPVersionNumber;
	
private:
	/** assembles a QStringList with all arguments to be given to PPPD
	 *  taking input from various sources, including the profile.
	 *  @param aProfile   the profile to take data from
	 *  @returns a QStringList with arguments
	 */
	QStringList assembleArgumentList(const Profile& aProfile) const;

	void setNewPPPState(PPPObserverInterface::PPPDState aNewState);

	/** runs the 'chat', i.e. a series of queries to prepare the modem
	 *  for the PPP connection - queries like ATD to dial in.
	 *  it will use the currently setup serialport - if that 
	 *  happens to be a TempSerial, that's not a problem :-)
	 *  @param aProfile   the profile to take data from
	 *  @returns true if successful
	 */
	bool chat(const Profile& aProfile) const;
	
	/** sends an actual line of the chat and waits for the answers
	 * @param aChatString the data to send
	 * @returns true if reply was "OK".
	 */
	bool sendChatString(const QString& aChatString) const;
	
private:
	/** copy constructor is private and not implemented
	 *  to prevent duplication of this class instance (EffC++ item27)
	 *  DO NOT IMPLEMENT THIS MEMBER
	 */
	PPPConnection(const PPPConnection& aRef);
	
	/** assign operator is private and not implemented
	 *  to prevent duplication of this class instance (EffC++ item27)
	 *  DO NOT IMPLEMENT THIS MEMBER
	 */
	PPPConnection& operator=(const PPPConnection& aRef);
	
	/** this nested class creates the separate execution thread that
	 *  will fork/exec into ppp - to make sure that the main thread
	 *  keeps running
	 *  FIXME: This might move into Runner later...
	 */ 
	class PPPThread : public QThread
	{
	public:
		PPPThread(const QStringList& anArgumentList);
		virtual ~PPPThread() {DEBUG2("~PPPThread()\n"); QThread::wait(); };
		
		/** I re-implemented this one because I suspect the QT one
		 *  (called finished())  to be misbehaving
		 *  @returns true if the thread has ended or hasn't started yet
		 */
		virtual bool isDone()	{ return !isRunning;};
		
		/// reimplemented from QThread to ensure that isDone is working
		virtual void start()	{isRunning=true; QThread::start();};
		
		const QStringList& getStdOut() const {return theRunner.getStdOutStringList();};
		const QStringList& getStdErr() const {return theRunner.getStdErrStringList();};

	protected:
		QStringList theArgumentList;

		/// this member contains the actual thread code
		virtual void run();
		
		int theReturnValue;
		bool isRunning;
		Runner theRunner;
		
		/// as long as thePPPThread exists, no info updates
		ConnectionInfoInhibitor theCII;
	}; // end-of-PPPThread declaration
	
	PPPThread* thePPPThreadPtr;
};

#endif /*PPPCONNECTION_H_*/
